#include "glob-store.h"
#include "glob-search.h"

enum
{
    FINISHED,
    LAST_SIGNAL
};

enum
{
    PROP_0,
    PROP_PATTERN,
    PROP_LAST
};

static GParamSpec *pspecs[PROP_LAST];

struct _GlobStorePrivate
{
    GlobSearch *search;
    gchar *pattern;
};

G_DEFINE_TYPE_WITH_PRIVATE (GlobStore, glob_store, GTK_TYPE_LIST_STORE)

static guint signals[LAST_SIGNAL] = { 0, };

static void
glob_store_set_property (GObject      *object,
                         guint         prop_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
    GlobStore *store = GLOB_STORE (object);

    switch (prop_id)
    {
        case PROP_PATTERN:
            glob_store_set_pattern (store, g_value_get_string (value));
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

static void
glob_store_get_property (GObject   *object,
                        guint       prop_id,
                        GValue     *value,
                        GParamSpec *pspec)
{
    GlobStore *store = GLOB_STORE (object);

    switch (prop_id)
    {
            case PROP_PATTERN:
            g_value_set_string (value, glob_store_get_pattern (store));
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

static void
glob_store_finalize (GObject *object)
{
    GlobStorePrivate *priv = GLOB_STORE (object)->priv;

    g_free (priv->pattern);

    G_OBJECT_CLASS (glob_store_parent_class)->finalize (object);
}

static void
glob_store_dispose (GObject *object)
{
    GlobStorePrivate *priv = GLOB_STORE (object)->priv;

    g_object_unref (priv->search);

    G_OBJECT_CLASS (glob_store_parent_class)->dispose (object);
}

static void
glob_store_class_init (GlobStoreClass *klass)
{
    GObjectClass *obj_class = G_OBJECT_CLASS (klass);

    obj_class->get_property = glob_store_get_property;
    obj_class->set_property = glob_store_set_property;
    obj_class->finalize = glob_store_finalize;
    obj_class->dispose = glob_store_dispose;

    pspecs[PROP_PATTERN] = g_param_spec_string ("pattern", "Pattern", "The pattern to match against",
                                                "",
                                                G_PARAM_READWRITE);
    g_object_class_install_property (obj_class, PROP_PATTERN, pspecs[PROP_PATTERN]);
                                     

    /**
     * GlobStore::finished:
     * @store: the object which received the signal
     *
     * Emitted when matching words has finished.
     */
    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (obj_class),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (GlobStoreClass, finished),
                      NULL, NULL, NULL,
                      G_TYPE_NONE, 1,
                      G_TYPE_INT);

    g_type_class_add_private (klass, sizeof (GlobStorePrivate));
}

static void
populate_finish (GObject *source_object,
                 GAsyncResult *res,
                 gpointer user_data)
{
    GtkListStore *store = GTK_LIST_STORE (user_data);
    gchar **list;
    gint matches;
    gint i;
    
    list = glob_search_get_matches_finish (GLOB_SEARCH (source_object), res, NULL);

    matches = g_strv_length (list);

    for (i = 0; i < matches; i++)
    {
        GtkTreeIter iter;

        gtk_list_store_append (store, &iter);
        gtk_list_store_set (store, &iter,
                            GLOB_STORE_TEXT, list[i],
                            -1);
    }

    g_signal_emit (store, signals[FINISHED], 0, matches);

    g_strfreev (list);
}

static void
populate_store (GObject    *object,
                GParamSpec *pspec)
{
    GtkListStore *store = GTK_LIST_STORE (object);
    GlobStorePrivate *priv = GLOB_STORE (store)->priv;

    gtk_list_store_clear (store);

    glob_search_get_matches_async (priv->search, priv->pattern, NULL, populate_finish, store);
}

static void
glob_store_init (GlobStore *store)
{
    GlobStorePrivate *priv;
    GType types[GLOB_STORE_COLUMNS] = { G_TYPE_STRING };

    priv = glob_store_get_instance_private (store);

    priv->pattern = g_strdup ("");
    priv->search = glob_search_dup ();

    gtk_list_store_set_column_types (GTK_LIST_STORE (store), GLOB_STORE_COLUMNS, types);

    g_signal_connect (store, "notify::pattern",
                      G_CALLBACK (populate_store), NULL);

    store->priv = priv;
}

/**
 * glob_store_new:
 *
 * Creates a #GlobStore. This can be used to update a #GtkTreeView with matches.
 *
 * Returns: a new #GlobStore
 */
GlobStore *
glob_store_new (void)
{
    return g_object_new (GLOB_TYPE_STORE, NULL);
}

/**
 * glob_store_get_pattern:
 * @store: a #GlobStore
 *
 * Gets the glob pattern currently used by the store to find matches against.
 *
 * Returns: the glob pattern
 */
const gchar *
glob_store_get_pattern (GlobStore *store)
{
    g_return_val_if_fail (GLOB_IS_STORE (store), NULL);

    return store->priv->pattern;
}

/**
 * glob_store_set_pattern:
 * @store: a #GlobStore
 * @glob: the new glob pattern
 *
 * Sets the new glob pattern for @store to use to find matches against.
 */
void
glob_store_set_pattern (GlobStore   *store,
                        const gchar *pattern)
{
    GlobStorePrivate *priv = store->priv;

    g_return_if_fail (GLOB_IS_STORE (store));
    g_return_if_fail (pattern != NULL);

    g_free (priv->pattern);
    priv->pattern = g_strdup (pattern);

    g_object_notify_by_pspec (G_OBJECT (store), pspecs[PROP_PATTERN]);
}
